-- "Classes"

Candidato = {nome_candidato = "", numero_votos = 0, partido = "", coliga = "", eleito = false, coligado = false}

Partido = {nome_partido = "", numero_votos = 0, numero_candidatos_eleitos = 0, coligado = false, lista_candidatos = {}}

Coligacao = {nome_coligacao = "", numero_votos = 0, numero_total_eleitos = 0, lista_partidos = {}}

-- Coligação
-- "Construtor"
function Coligacao:new (o)
  o = o or {}
  setmetatable(o, self)
  self.__index = self
  return o
end

-- Retorno do Nome da Coligação
function Coligacao:getName()
  return self.nome_coligacao
end

-- Aumenta a Quantidade de Votos de Coligação
function Coligacao:atualizaVotos(valor)
  valor = valor or 0
  self.numero_votos = self.numero_votos + valor
end

-- Aumenta a Quantidade de Candidatos Eleitos na Coligação
function Coligacao:incrementaEleitos()
  self.numero_total_eleitos = self.numero_total_eleitos + 1
end

-- Imprime a Coligação como especificado
function Coligacao:imprime(id)
  if (self.numero_total_eleitos >= 2) then
    print(id .. " - Coligação: " .. self.nome_coligacao .. ", " .. self.numero_votos .. " votos, " .. self.numero_total_eleitos .. " candidatos eleitos")
  else
    print(id .. " - Coligação: " .. self.nome_coligacao .. ", " .. self.numero_votos .. " votos, " .. self.numero_total_eleitos .. " candidato eleito")
  end
end

-- Candidato
-- "Construtor"
function Candidato:new (o)
  o = o or {}
  setmetatable(o, self)
  self.__index = self
  return o
end

-- Imprime o Candidato como especificado
function Candidato:imprime (id)
  if self.coligado then
    print(id.. " - " .. self.nome_candidato .. " (" ..self.partido.. ", " ..self.numero_votos.." votos)" .. " - Coligação: " .. self.coliga)
  else
    print(id.. " - " .. self.nome_candidato .. " (" ..self.partido.. ", " ..self.numero_votos.." votos)")
  end
end

-- Partido
-- "Construtor"
function Partido:new (o)
  o = o or {}
  setmetatable(o, self)
  self.__index = self
  return o
end

-- Retorna o Nome do Partido
function Partido:getName()
  return self.nome_partido
end

-- Aumenta a Quantidade de Votos de Partido
function Partido:atualizaVotos(valor)
  valor = valor or 0
  self.numero_votos = self.numero_votos + valor
end

-- Aumenta a Quantidade de Candidatos Eleitos no Partido
function Partido:incrementaEleitos()
  self.numero_candidatos_eleitos = self.numero_candidatos_eleitos + 1
end

-- Adiciona o Candidato na lista de Candidatos do Partido
function Partido:incrementaCandidatos(candidato)
  self.candidatos = self.candidatos or {}
  table.insert(self.candidatos, candidato)
  if candidato.eleito then
    self:incrementaEleitos()
  end
end

-- Imprime Partido como especificado
function Partido:imprime(id)
  if self.numero_candidatos_eleitos >= 2 then
    print(id .. " - " .. self.nome_partido .. ", " .. self.numero_votos .. " votos, " .. self.numero_candidatos_eleitos .. " candidatos eleitos")
  else
    print(id .. " - " .. self.nome_partido .. ", " .. self.numero_votos .. " votos, " .. self.numero_candidatos_eleitos .. " candidato eleito")
  end
end


-- Inicio da Leitura de Arquivos

-- Verifica se o arquivo existe
function file_exists(file)
  local f = io.open(file, "rb")
  if f then f:close() end
  return f ~= nil
end

-- Faz a leitura de todas as linhas
function lines_from(file)
  if not file_exists(file) then return {} end
  lines = {}
  for line in io.lines(file) do
    lines[#lines + 1] = line
  end
  return lines
end

-- Função para Split
-- Função retirada do StackOverFlow
split = function(s, pattern, maxsplit)
  local pattern = pattern or ' '
  local maxsplit = maxsplit or -1
  local s = s
  local t = {}
  local patsz = #pattern
  while maxsplit ~= 0 do
    local curpos = 1
    local found = string.find(s, pattern)

    if found ~= nil then
      table.insert(t, string.sub(s, curpos, found - 1))
      curpos = found + patsz
      s = string.sub(s, curpos)
    else
      table.insert(t, string.sub(s, curpos))
      break
    end
    maxsplit = maxsplit - 1

    if maxsplit == 0 then
      table.insert(t, string.sub(s, curpos - patsz - 1))
    end
  end
  return t
end

-- Verifica se o Partido se encontra na Lista de Partidos
function verificaPartidoNaLista(nome)
  for k = 1,#lista_partidos do
    if lista_partidos[k]:getName() == nome then
      return true
    end
  end
  return false
end

-- Retorna a posição onde o Partido se encontra na Lista de Partidos
function posicaoNaListaDePartidos(nome)
  for k = 1,#lista_partidos do
    if lista_partidos[k]:getName() == nome then
      return k
    end
  end
  return 0
end

-- Verifica se a Coligação se encontra na Lista de Coligações
function verificaColigacaoNaLista(nome)
  for k = 1,#lista_coligacoes do
    if lista_coligacoes[k]:getName() == nome then
      return true
    end
  end
  return false
end

-- Retorna a posição onde a Coligação se encontra na Lista de Coligações
function posicaoNaListaDeColigacoes(nome)
  for k = 1,#lista_coligacoes do
    if lista_coligacoes[k]:getName() == nome then
      return k
    end
  end
  return 0
end

-- Faz a criação dos Objetos (Candidatos, Partidos e Coligações) com as linhas lidas do Arquivo
function init(lines)
  for k in pairs(lines) do
    if k > 1 then
      local linha = split(lines[k], ";")
      local votos = tonumber(tostring(linha[5]:gsub('%.', '')))
      local el = false
      local cand_nome = linha[3]
      local part_nome = ""
      local colig_nome = ""
      local cand_coligado = false

      if string.find(linha[1], "*") then
        vagas = vagas + 1
        el = true
      end

      if string.find(linha[4], "-") then
        local part_colig = split(linha[4], "-")
        part_nome = string.sub(part_colig[1], 1, -2)
        colig_nome = string.sub(part_colig[2], 2, -1)
        cand_coligado = true
      else
        part_nome = linha[4]
        colig_nome = linha[4]
      end

      local candidato = Candidato:new{nome_candidato = cand_nome, numero_votos = votos, partido = part_nome, coliga = colig_nome, eleito = el, coligado = cand_coligado}
      local partido = Partido:new{nome_partido = part_nome}
      -- Verifica se tem PARTIDO
      if verificaPartidoNaLista(part_nome) then
        local value = posicaoNaListaDePartidos(part_nome)
        lista_partidos[value]:atualizaVotos(votos)
        lista_partidos[value]:incrementaCandidatos(candidato)
      else
        table.insert(lista_partidos, partido)
        lista_partidos[#lista_partidos]:atualizaVotos(votos)
        lista_partidos[#lista_partidos]:incrementaCandidatos(candidato)
      end
      -- Verifica se tem Coligacao
      if verificaColigacaoNaLista(colig_nome) then
        local value = posicaoNaListaDeColigacoes(colig_nome)
        table.insert(lista_coligacoes[value].lista_partidos, partido)
        if el then
          lista_coligacoes[value]:incrementaEleitos()
        end
        lista_coligacoes[value]:atualizaVotos(votos)
      else
        table.insert(lista_coligacoes, Coligacao:new{nome_coligacao = colig_nome, numero_votos = votos})
        table.insert(lista_coligacoes[#lista_coligacoes].lista_partidos, partido)
        if el then
          lista_coligacoes[#lista_coligacoes]:incrementaEleitos()
        end
      end

      table.insert(lista_candidatos, candidato)
    end
  end
end


-- Realiza a Impressão conforme especificado
function imprime()
  local i = 1
  local votos = 0
  print("Número de vagas: " ..vagas.. "\n\n" .. "Vereadores eleitos:")
  for k in pairs(lista_candidatos) do
    if lista_candidatos[k].eleito then
      lista_candidatos[k]:imprime(i)
      i = i + 1
    end
    votos = lista_candidatos[k].numero_votos  + votos
  end
  print("\nCandidatos mais votados (em ordem decrescente de votação e respeitando número de vagas):")
  for k = 1,vagas do
    lista_candidatos[k]:imprime(k)
  end
  i = 1
  print("\nTeriam sido eleitos se a votação fosse majoritária, e não foram eleitos:\n(com sua posição no ranking de mais votados)")
  for k in pairs(lista_candidatos) do
    if not(lista_candidatos[k].eleito) then
      lista_candidatos[k]:imprime(i)
    end
    i = i + 1
    if i > vagas then
      break
    end
  end
  print("\nEleitos, que se beneficiaram do sistema proporcional:\n(com sua posição no ranking de mais votados)")
  auxEleitos = 0
  i = 1
  while auxEleitos ~= vagas do
    if lista_candidatos[i].eleito then
      if i >= vagas then
        lista_candidatos[i]:imprime(i)
      end
      auxEleitos = auxEleitos + 1
    end
    i = i + 1
  end
  print("\nVotação (nominal) das coligações e número de candidatos eleitos:")
  i = 1
  for k in pairs(lista_coligacoes) do
    lista_coligacoes[k]:imprime(i)
    i = i + 1
  end
  	i = 1
  print("\nVotação (nominal) dos partidos e número de candidatos eleitos:")
  for k in pairs(lista_partidos) do
    lista_partidos[k]:imprime(i)
    i = i + 1
  end

  print("\nTotal de votos nominais: "..votos)
end


-- Função de comparação para o Table Sort
function compare(a,b)
  return a.numero_votos > b.numero_votos
end

-- Script Principal
local entrada = 'divulga.csv' -- Abertura do arquivo de entrada
local lines = lines_from(entrada)
lista_candidatos = {}
lista_partidos = {}
lista_coligacoes = {}
vagas = 0
init(lines)
table.sort(lista_candidatos, compare)
table.sort(lista_partidos, compare)
table.sort(lista_coligacoes, compare)
imprime()
